#include <cmath>
#include <iostream>
#include <GL/glew.h>
#include <GL/freeglut.h>
#include <GL/freeglut_ext.h>
#include <glm/glm.hpp>
#include <vector>
#include "moglclass.h"
#include "shaders.h"
#include "vertexBufferObject.h"

using namespace moglclass;

#ifndef M_PI
#define M_PI 3.141592
#endif
#define MOGL_STEP 7
#define MOGL_THICKNESS 3

#define NUMSHADERS 2
Shader shShaders[NUMSHADERS];
Program spMain;

GLuint triangleVAO;
VertexBufferObject triangleVBO;
int triangleNumVertices = 3;

GLuint squareVAO;
VertexBufferObject squareVBO;
int squareNumVertices = 6;

GLuint circleVAO;
VertexBufferObject circleVBO;
int circleNumVertices = MOGL_STEP;

GLuint indexedSquareVAO;
std::vector<glm::vec4> squareVertices; // Square vertices
std::vector<glm::vec4> squareColors; // Square colors
std::vector<int> squareIndices; // Indices into squareVerts[] buffer for 2 triangles

int g_geomode = 0; // 0=triangle, 1=square, 2=circle, 3=indexedSquare
int g_colormode = 0;
GLenum g_primitivemode = GL_POINTS;
bool g_wiremode = false;

////////////////////////////////////////////////////////////////////////////////////////////////////
// set color triangle
void setTriangleData()
{
	// VAO
	glGenVertexArrays(1, &triangleVAO);
	glBindVertexArray(triangleVAO);

	glm::vec4 v1 = glm::vec4(-0.75f, -0.75f, 0.0f, 1.0f);
	glm::vec4 c1 = glm::vec4(1.0f, 0.0f, 0.0f, 1.0f);
	glm::vec4 v2 = glm::vec4(0.75f, -0.75f, 0.0f, 1.0f);
	glm::vec4 c2 = glm::vec4(0.0f, 1.0f, 0.0f, 1.0f);
	glm::vec4 v3 = glm::vec4(0.75f, 0.75f, 0.0f, 1.0f);
	glm::vec4 c3 = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);

	// VBO
	triangleVBO.createVBO();
	triangleVBO.bindVBO();
	triangleVBO.addData(&v1, sizeof(glm::vec4)); // vertex1 position
	triangleVBO.addData(&c1, sizeof(glm::vec4));		// vertex1 color
	triangleVBO.addData(&v2, sizeof(glm::vec4));	// vertex2 position
	triangleVBO.addData(&c2, sizeof(glm::vec4));		// vertex2 color
	triangleVBO.addData(&v3, sizeof(glm::vec4));	// vertex3 position
	triangleVBO.addData(&c3, sizeof(glm::vec4));		// vertex3 color
	triangleVBO.uploadDataToGPU(GL_STATIC_DRAW);

	// vertex attrib
	int iDataStride = 2 * sizeof(glm::vec4); // vertex & color
	int iDataOffset = 0;
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, iDataStride, (void*)iDataOffset);
	iDataOffset += sizeof(glm::vec4);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, iDataStride, (void*)iDataOffset);
}

void setSquareData()
{
	// VAO
	glGenVertexArrays(1, &squareVAO);
	glBindVertexArray(squareVAO);

	glm::vec4 v1 = glm::vec4(-0.75f, -0.75f, 0.0f, 1.0f);
	glm::vec4 v2 = glm::vec4(0.75f, -0.75f, 0.0f, 1.0f);
	glm::vec4 v3 = glm::vec4(0.75f, 0.75f, 0.0f, 1.0f);
	glm::vec4 v4 = glm::vec4(-0.75f, 0.75f, 0.0f, 1.0f);
	glm::vec4 c1 = glm::vec4(1.0f, 0.0f, 0.0f, 1.0f);
	glm::vec4 c2 = glm::vec4(0.0f, 1.0f, 0.0f, 1.0f);
	glm::vec4 c3 = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
	glm::vec4 c4 = glm::vec4(1.0f, 0.0f, 1.0f, 1.0f);

	// VBO
	squareVBO.createVBO();
	squareVBO.bindVBO();
	//face1
	squareVBO.addData(&v1, sizeof(glm::vec4)); // vertex1 position
	squareVBO.addData(&c1, sizeof(glm::vec4));		// vertex1 color
	squareVBO.addData(&v2, sizeof(glm::vec4)); // vertex2 position
	squareVBO.addData(&c2, sizeof(glm::vec4));		// vertex2 color
	squareVBO.addData(&v3, sizeof(glm::vec4)); // vertex3 position
	squareVBO.addData(&c3, sizeof(glm::vec4));		// vertex3 color
	// face2
	squareVBO.addData(&v1, sizeof(glm::vec4)); // vertex1 position
	squareVBO.addData(&c1, sizeof(glm::vec4));		// vertex1 color
	squareVBO.addData(&v3, sizeof(glm::vec4)); // vertex3 position
	squareVBO.addData(&c3, sizeof(glm::vec4));		// vertex3 color
	squareVBO.addData(&v4, sizeof(glm::vec4)); // vertex4 position
	squareVBO.addData(&c4, sizeof(glm::vec4));		// vertex4 color
	squareVBO.uploadDataToGPU(GL_STATIC_DRAW);

	// vertex attrib
	int iDataStride = 2 * sizeof(glm::vec4); // vertex & color
	int iDataOffset = 0;
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, iDataStride, (void*)iDataOffset);
	iDataOffset += sizeof(glm::vec4);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, iDataStride, (void*)iDataOffset);

	glBindVertexArray(0);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// set color circle
void setCircleData(float radius, int step)
{
	// VAO
	glGenVertexArrays(1, &circleVAO);
	glBindVertexArray(circleVAO);

	// VBO
	circleVBO.createVBO();
	circleVBO.bindVBO();
	glm::vec4 vertex(0.0f, 0.0f, 0.0f, 1.0f);
	glm::vec4 color(1.0f, 0.0f, 0.0f, 1.0f);
	float theta = (float) (2*M_PI/step);
	for (int i=0; i<step; i++) 
	{
		vertex[0] = radius * cos(theta * i);
		vertex[1] = radius * sin(theta * i);
		circleVBO.addData(&vertex[0], sizeof(glm::vec4)); // vertex position
		circleVBO.addData(&color[0], sizeof(glm::vec4)); // vertex color
	}
	circleNumVertices = step;
	circleVBO.uploadDataToGPU(GL_STATIC_DRAW);

	// vertex attrib
	int iDataStride = 2 * sizeof(glm::vec4); // vertex & color
	int iDataOffset = 0;
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, iDataStride, (void*)iDataOffset);
	iDataOffset += sizeof(glm::vec4);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, iDataStride, (void*)iDataOffset);
}

void setIndexedSquareData()
{
	// Vertice Positions
	squareVertices.push_back(glm::vec4(-0.75f, -0.75f, 0.0f, 1.0f));
	squareVertices.push_back(glm::vec4(0.75f, -0.75f, 0.0f, 1.0f));
	squareVertices.push_back(glm::vec4(0.75f, 0.75f, 0.0f, 1.0f));
	squareVertices.push_back(glm::vec4(-0.75f, 0.75f, 0.0f, 1.0f));
	// Vertices Colors
	squareColors.push_back(glm::vec4(1.0f, 0.0f, 0.0f, 1.0f));
	squareColors.push_back(glm::vec4(0.0f, 1.0f, 0.0f, 1.0f));
	squareColors.push_back(glm::vec4(0.0f, 0.0f, 1.0f, 1.0f));
	squareColors.push_back(glm::vec4(0.0f, 0.0f, 0.0f, 1.0f));
	// Vertice Indices
	squareIndices.push_back(0);
	squareIndices.push_back(1);
	squareIndices.push_back(2);
	squareIndices.push_back(0);
	squareIndices.push_back(2);
	squareIndices.push_back(3);

	// VAO
	GLuint vao;
	glGenVertexArrays(1, &indexedSquareVAO);
	glBindVertexArray(indexedSquareVAO);

	// VBO
	GLuint vbo[2], ibo;
	glGenBuffers(2, &vbo[0]);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
	glBufferData(GL_ARRAY_BUFFER, 4 * sizeof(glm::vec4), &squareVertices[0], GL_STATIC_DRAW);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
	glBufferData(GL_ARRAY_BUFFER, 4 * sizeof(glm::vec4), &squareColors[0], GL_STATIC_DRAW);
	glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(1);

	// IBO
	glGenBuffers(1, &ibo);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, 6 * sizeof(int), &squareIndices[0], GL_STATIC_DRAW);
	//glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);

	glBindVertexArray(0);
	squareVertices.clear();
	squareColors.clear();
}

void init( void )
{
    // Load shaders and use the resulting shader program
	spMain.createProgram();
	shShaders[0].loadShader("simple2.vs", GL_VERTEX_SHADER);
	shShaders[1].loadShader("simple2.fs", GL_FRAGMENT_SHADER);
	spMain.attachShaderToProgram(&shShaders[0]);
	spMain.attachShaderToProgram(&shShaders[1]);
	spMain.linkProgram();
	spMain.useProgram();
	
	// Set data
	setTriangleData();
	setSquareData();
	setCircleData(0.75f, MOGL_STEP);
	setIndexedSquareData();

	glClearColor( 1.0, 1.0, 1.0, 1.0 ); // white background
}

void drawCircle()
{
	// circle
	glBindVertexArray(circleVAO);
	if (g_colormode == 0)
		glVertexAttrib3f((GLuint)1, 1, 0, 0); // set constant color attribute (red)
	else if (g_colormode == 1)
		glVertexAttrib3f((GLuint)1, 0, 1, 0); // set constant color attribute (green)
	else if (g_colormode == 2)
		glVertexAttrib3f((GLuint)1, 0, 0, 1); // set constant color attribute (blue)

	// set drawmode
	if (g_wiremode)
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);  // draw wireframe 
	else
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);	// draw filled 

	// primitive type
	if (g_primitivemode == GL_POINTS) // 0
	{
		glPointSize(MOGL_THICKNESS);
		glDrawArrays(GL_POINTS, 0, circleNumVertices);       // draw points
		glPointSize(1);
	}
	else if (g_primitivemode == GL_LINES) // 1
	{
		glLineWidth(MOGL_THICKNESS);
		glDrawArrays(GL_LINES, 0, circleNumVertices);       // draw lines
		glLineWidth(1);
	}
	else if (g_primitivemode == GL_LINE_LOOP) // 2
	{
		glLineWidth(MOGL_THICKNESS);
		glDrawArrays(GL_LINE_LOOP, 0, circleNumVertices);       // draw line loop
		glLineWidth(1);
	}
	else if (g_primitivemode == GL_LINE_STRIP) // 3
	{
		glLineWidth(MOGL_THICKNESS);
		glDrawArrays(GL_LINE_STRIP, 0, circleNumVertices);       // draw line strip
		glLineWidth(1);
	}
	else if (g_primitivemode == GL_TRIANGLES) // 4
	{
		glDrawArrays(GL_TRIANGLES, 0, circleNumVertices);       // draw triangles
	}
	else if (g_primitivemode == GL_TRIANGLE_STRIP) // 5
	{
		glDrawArrays(GL_TRIANGLE_STRIP, 0, circleNumVertices);       // draw triangle strip
	}
	else if (g_primitivemode == GL_TRIANGLE_FAN) // 6
	{
		glDrawArrays(GL_TRIANGLE_FAN, 0, circleNumVertices);       // draw triangle fan
	}
	else if (g_primitivemode == GL_POLYGON) // 9
	{
		glDrawArrays(GL_POLYGON, 0, circleNumVertices);       // draw polygon
	}

	glBindVertexArray(0);
	glDisableVertexAttribArray(0);
}

void drawTriangle()
{
	// triangle
	glBindVertexArray(triangleVAO);
	if (g_wiremode)
		glDrawArrays(GL_LINE_LOOP, 0, triangleNumVertices);       // draw wireframe
	else
		glDrawArrays(GL_TRIANGLES, 0, triangleNumVertices);       // draw triangle
	glBindVertexArray(0);
	glDisableVertexAttribArray(0);
}

void drawIndexedSquare()
{
	// square
	glBindVertexArray(indexedSquareVAO);
	if (g_wiremode)
		glDrawElements(GL_LINE_LOOP, 6, GL_UNSIGNED_INT, 0);       // draw wireframe
	else
		glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, 0);       // draw filled
	glBindVertexArray(0);
	glDisableVertexAttribArray(0);
}

void drawSquare()
{
	// square
	glBindVertexArray(squareVAO);
	if (g_wiremode)
		glDrawArrays(GL_LINE_LOOP, 0, squareNumVertices);       // draw wireframe
	else
		glDrawArrays(GL_TRIANGLES, 0, squareNumVertices);       // draw filled
	glBindVertexArray(0);
	glDisableVertexAttribArray(0);
}

void display( void )
{
    glClear( GL_COLOR_BUFFER_BIT );     // clear the window

	switch (g_geomode) {
	case 0:
		drawTriangle();
		break;
	case 1:
		drawSquare();
		break;
	case 2:
		drawCircle();
		break;
	case 3:
		drawIndexedSquare();
		break;
	}

	glFlush();
}

void keyboard( unsigned char key, int x, int y )
{
	switch (key)  
	{    
	case 0x1B:    
	case 'q':    
	case 'Q':      
		exit(0);     
	case ' ':
		g_primitivemode++;
		if (g_primitivemode >= GL_POLYGON) g_primitivemode = GL_POINTS;
		else if (g_primitivemode > GL_TRIANGLE_FAN) g_primitivemode = GL_POLYGON;
		else g_primitivemode = g_primitivemode % 7;
		printf("g_primitivemode=%d\n", g_primitivemode);
		break;
	case 'c':
		g_colormode = (g_colormode+1) % 3;
		printf("g_colormode=%d\n", g_colormode);
		break;  
	case 'g':
		g_geomode = (g_geomode + 1) % 4;
		printf("g_geomode=%d\n", g_geomode);
		break;
	case 'w':
		g_wiremode = !g_wiremode;
		printf("g_wiremode=%d\n", g_wiremode);
		break;
	}
	glutPostRedisplay();
}

void reshape(int w, int h)
{
	glViewport(0, 0, w, h);
}

void cleanup()
{
	spMain.deleteProgram();
}

int main( int argc, char **argv )
{
	glutInit(&argc, argv);
    glutInitDisplayMode( GLUT_RGBA );
    glutInitWindowSize( 512, 512 );
    glutCreateWindow(argv[0]);
    
	glewInit();
    init();
    glutDisplayFunc( display );
    glutKeyboardFunc( keyboard );
	glutReshapeFunc( reshape );
	glutMainLoop();

	cleanup();
    return 0;
}
